# import sys; sys.path.insert(0, "/home/jborst/tmp/utsi")
#
#
# from utsi import DataManager
# dm = DataManager()
#
# dm.read_conll("/disk1/users/jborst/Data/Test/NER/CoNLL-2003/de/BIOES/test_2c_small.txt", purpose="test",columns=["token","ner"])
# # dm.read_json("/disk1/users/jborst/Data/Test/NER/Wiki/test_comma.json", purpose="test")
#

# from raziel.Raziel import Raziel
# bb = Raziel(dm)
# bb.schemetransformer("ner","noprefix")
# #
# # with open("/disk1/users/jborst/Data/Test/NER/Wiki/tagset", "r") as tagsetInput:
# #     tagset = tagsetInput.read().split("\n")[:-1] + ["/education"]
# # bb.setTagset(tagset, scheme="noprefix")
# # bb.language = "de"
# # bb.sequence_length = 140
# # bb.character_length = 60
# # # bb.setTagset(tagset=["LOC", "PER", "ORG", "MISC"], scheme="iobes")
# #
# # traindata = bb.chain({
# #     # "chars": ("token", "character", "index", None),
# #     # "w2v":          ("token", "custom", "embedding", "/disk1/users/jborst/Data/Pretrained Word Embeddings/german/cc.de.300.10K.vec"), # only act on something as file
# #     #"elmo":         ("token", "elmo", "embedding", None),
# #     "mentions":    ("mentions",   "multilabel", "index", "sparse" ),
# #     # "fasttext":     ("token", "fasttext", "embedding", "/disk1/users/jborst/Data/Pretrained Word Embeddings/german/cc.de.300.bin")
# #     # "ner_embedding": ("ner", "custom", "embedding", "/disk1/users/jborst/Data/Pretrained Word Embeddings/test.txt")
# # })
# # for data in bb.transform_iter(10,"test"):
# #     print(data)
# #     break
#
#
import sys;
sys.path.insert(0,"/home/jborst/tmp/utsi")
sys.path.insert(0,"/home/jborst/tmp/raziel")
sys.path.insert(0,"/home/jborst/tmp/ptagger")
from tqdm import tqdm
import numpy as np
import itertools
from utsi import DataManager
from raziel.Raziel import Raziel
dm = DataManager()
bb = Raziel(dm)
#
# scheme = "iobes"
#
# bb.data_manager.read_conll("/disk1/users/jborst/Data/Test/NER/CoNLL-2003/de/BIOES/train.txt", purpose="embed")
# bb.data_manager.read_conll("/disk1/users/jborst/Data/Test/NER/CoNLL-2003/de/BIOES/valid.txt", purpose="embed")
# bb.data_manager.read_conll("/disk1/users/jborst/Data/Test/NER/CoNLL-2003/de/BIOES/test.txt", purpose="test")
# tagset = ["PER", "ORG", "LOC", "MISC"]
# bb.setTagset(tagset, scheme=scheme)
# # bb.schemetransformer("ner", scheme)
# bb.max_sequence_length = 10
#
# bb.chain({
#    "embeddings": ("token", "custom", "embedding","/disk1/users/jborst/Data/Pretrained Word Embeddings/german/cc.de.300.10K.vec"),
#     "tags": ("ner", "label", "index", "sparse")
# })
#
# data = bb.transform("test")

dm = DataManager()
dm.read_conll("/home/jborst/semEval_task1.conll",columns=["token", "pos", "lposition","rposition", "relation"])
sample = dm.random(100)
from utsi.tasks.implementations_ethds3 import ETH3DS
p=ETH3DS()
p.initialize()
p.train(dm.get_data(sample))
p.run(data=dm.get_data())


import numpy as np
from raziel.vocab import Vocabulary
class plugin_custom:
    def __init__(self, level, option, **kwargs):
        self.vocab= Vocabulary()
        self.name = "custom"
        self.level = level
        self.word_length = kwargs["word_length"]
        self.sentence_length = kwargs["sentence_length"]
        assert level in ["index", "embedding"], "only 'index' and 'embedding' supported"
        self.option=option
        assert option is not None, "Custom needs a path to a vector file"
        self.vocab.setWordEmbedding(option)

        if level == "index":
            self.shape = (self.sentence_length, )
        elif level == "embedding":
            self.shape = (self.sentence_length, self.vocab.wordEmbeddingSize)

        print("custom embedding loaded")


    def getEmbedding(self, w, level="embedding"):
        index = self.vocab.getWordIndex(w)

        if self.level == "index":
            index_array=np.full((1,self.sentence_length), 0., dtype="int32")
            index_array[0,:min(self.sentence_length,len(index))] = index
            return index_array
        else:
            embedded = [self.vocab.wordEmbedding.vectors[i] for i in index]
            sentence_embedding = np.full((1, self.sentence_length, self.vocab.wordEmbeddingSize), 0., dtype="float32")
            sentence_embedding[0, :min(len(embedded), self.sentence_length),:] = np.array(embedded[:min(len(embedded), self.sentence_length)])
            return (sentence_embedding)

    def getDim(self):
        return (self.vocab.wordEmbeddingSize,)

# kwargs = {
#             "sentence_length": 100,
#             "word_length": 60
#         }
# p = plugin_custom(level="embedding",option = "/disk1/users/jborst/Data/Pretrained Word Embeddings/glove.840B.300d_10k.w2vformat.txt", **kwargs)
# e = p.getEmbedding(["This","is","a","Sentence","."])
# e.shape


import numpy as np
# import fastText as ft
from gensim.models.fasttext import FastText as ft
class plugin_fasttext:
    def __init__(self, level, option, **kwargs):
        self.name = "custom"
        self.level = level
        self.sentence_length = kwargs["sentence_length"]
        assert level in ["embedding"], "only 'embedding' supported"
        self.option = option
        assert option is not None, "Custom needs a path to a vector file"

        self.embedding = ft.load_fasttext_format(option)

        self.shape = (self.sentence_length, self.embedding.vector_size)
        print("Fasttext Plugin Loaded")

    def getEmbedding(self,w):
        e = np.full((len(w), self.embedding.vector_size),0)
        try:
            e = np.array([self.embedding.wv[x] for x in w])
        except:
            Warning("Fasttext could not convert: " + " ".join(w))
        for i in np.where(~e.any(axis=1))[0]:
            e[i] = e[i] + 1e-30

        array = np.full((1, self.sentence_length, self.embedding.vector_size), 0.,dtype="float32")
        array[0,:min(self.sentence_length, e.shape[0]),:] = e[:min(self.sentence_length, e.shape[0]),:]
        return array
    def getDim(self):
        return (self.embedding.get_dimension(),)



from flair.embeddings import FlairEmbeddings # init embedding
from flair.data import Sentence
from flair.embeddings import StackedEmbeddings

import numpy as np
class plugin_flair:
    def __init__(self, level, option,**kwargs):
        assert level=="embedding", "Only level embedding supported"
        self.sentence_length = kwargs["sentence_length"]
        # init standard GloVe embedding
        self.embeddings = []
        if option is not None:
            assert isinstance(option, list), "Flair options must be a list of embeddings"
            for emb in option:
                self.embeddings.append(FlairEmbeddings(emb))


        # init standard character embeddings
        self.stacked_embeddings = StackedEmbeddings(embeddings=self.embeddings)
        print("Flair Plugin Loaded (Contextualised String Embedding)...")

        self.shape= (self.sentence_length, self.stacked_embeddings.embedding_length)

    def load(self,path):
        print("Flair loaded")

    def getEmbedding(self,w):
        embeddedSentence = Sentence(" ".join(w))
        self.stacked_embeddings.embed(embeddedSentence)
        embedArray = np.array([e.embedding.tolist() for e in embeddedSentence])
        if len(embedArray.shape)==1:
            embedArray = embedArray[np.newaxis,:]
        array = np.full((1, )+self.shape, 0., dtype="float32")
        array[0, :min(self.sentence_length, embedArray.shape[0]),:] = embedArray
        return array

    def getDim(self):
        return self.shape
#
#
# p = plugin_flair("embedding",option=["german-forward","german-backward"], sentence_length=140)
# p.getEmbedding(["hello","world"]).shape


import numpy as np
from allennlp.commands.elmo import ElmoEmbedder, \
    DEFAULT_OPTIONS_FILE, DEFAULT_WEIGHT_FILE
import os
class plugin_elmo:

    def concatenate(self, elmo_embeddings):
        """Helper for elmo layer concatenation."""

        return [np.concatenate(elmo_embeddings[:, i, :]) for i in range(elmo_embeddings.shape[1])]

    def __init__(self, level, option, **kwargs):
        self.sentence_length = kwargs["sentence_length"]
        self.level = level
        assert level in ["embedding"], "Index not reasonable for elmo plugin"
        self.shape = (self.sentence_length, 3072)
        # use 5.5B default elmo model
        self.options_file = DEFAULT_OPTIONS_FILE
        self.weight_file = DEFAULT_WEIGHT_FILE

        if option is not None:
            if os.path.isdir(option):
                json = [x for x in os.listdir(option) if x.endswith("json")]
                hdf5 = [x for x in os.listdir(option) if x.endswith("hdf5")]
                if len(json) ==1 and len(hdf5) == 1:
                    self.options_file = option+"/"+json[0]
                    self.weight_file = option+"/"+hdf5[0]
            elif option == 'small':
                self.options_file = 'https://s3-us-west-2.amazonaws.com/allennlp/models/elmo/2x1024_128_2048cnn_1xhighway/elmo_2x1024_128_2048cnn_1xhighway_options.json'
                self.weight_file = 'https://s3-us-west-2.amazonaws.com/allennlp/models/elmo/2x1024_128_2048cnn_1xhighway/elmo_2x1024_128_2048cnn_1xhighway_weights.hdf5'
            elif option == 'medium':
                self.options_file = 'https://s3-us-west-2.amazonaws.com/allennlp/models/elmo/2x2048_256_2048cnn_1xhighway/elmo_2x2048_256_2048cnn_1xhighway_options.json'
                self.weight_file = 'https://s3-us-west-2.amazonaws.com/allennlp/models/elmo/2x2048_256_2048cnn_1xhighway/elmo_2x2048_256_2048cnn_1xhighway_weights.hdf5'
            elif option == 'pt' or option == 'portuguese':
                self.options_file = 'https://s3-us-west-2.amazonaws.com/allennlp/models/elmo/contributed/pt/elmo_pt_options.json'
                self.weight_file = 'https://s3-us-west-2.amazonaws.com/allennlp/models/elmo/contributed/pt/elmo_pt_weights.hdf5'



        # INITIALIZATIONS ######################################################################

        # intialize elmo embedding
        self.elmo_embedder = ElmoEmbedder(self.options_file, self.weight_file)
        print("ELMo Plugin Loaded")
        _tmp = self.elmo_embedder.embed_sentence(["hallo"]).shape
        self.shape = (self.sentence_length,_tmp[0],_tmp[2])

    def load(self, path):
        print("ELMo loaded")
    def getEmbedding(self,w):
        elmo_emb = self.elmo_embedder.embed_sentence(w if w != [] else ['.'])
        elmo_emb = np.transpose(elmo_emb, [1,0,2])

        sentence_embedding = np.full((1,)+self.shape, 0., dtype="float32")
        sentence_embedding[0,:min(elmo_emb.shape[0], self.sentence_length)] = elmo_emb
        return(sentence_embedding)
    def getDim(self):
        return self.shape

#
# p = plugin_elmo(level="embedding",option = "/disk1/users/jborst/Data/Embeddings/elmo_german_de-ws", sentence_length=140)
# e = p.getEmbedding(["This","is","a","Sentence","."])
# e.shape


from flair.embeddings import BertEmbeddings # init embedding
from flair.data import Sentence
from flair.embeddings import StackedEmbeddings
from flair.embeddings import WordEmbeddings, CharacterEmbeddings

import numpy as np
class plugin_bert:
    def __init__(self, level, option,**kwargs):
        assert level=="embedding", "Only level embedding supported"
        self.sentence_length = kwargs["sentence_length"]
        # init standard GloVe embedding
        if option is None:
            self.embedding = BertEmbeddings()
        else:
            self.embedding = BertEmbeddings(option)

        # init standard character embeddings
        print("Bert Plugin Loaded (Contextualised String Embedding)...")

        self.shape= (self.sentence_length, self.embedding.embedding_length)

    def load(self,path):
        print("Flair loaded")

    def getEmbedding(self,w):
        embeddedSentence = Sentence(" ".join(w))
        self.embedding.embed(embeddedSentence)
        embedArray = np.array([e.embedding.tolist() for e in embeddedSentence])
        if len(embedArray.shape)==1:
            embedArray = embedArray[np.newaxis,:]
        array = np.full((1,) + self.shape, 0., dtype="float32")
        array[0, :min(self.sentence_length, embedArray.shape[0]),
        :] = embedArray
        return array

    def getDim(self):
        return self.shape


# p = plugin_bert("embedding", option="bert-base-cased", sentence_length=140)
# p.getEmbedding(["hallo","welt"])


from keras.layers import Layer
import keras.backend as K


class WeightedMean(Layer):
    def __init__(self, **kwargs):
        self.output_dim = 3
        super(WeightedMean, self).__init__(**kwargs)

    def build(self, input_shape):
        # Create a trainable weight variable for this layer.
        print(input_shape)
        self.kernel = self.add_weight(name='kernel',
                                      shape=(input_shape[-2],),
                                      initializer='uniform',
                                      trainable=True)
        super(WeightedMean, self).build(input_shape)  # Be sure to call this at the end

    def call(self, x):
        # return K.concatenate([K.repeat_elements(K.expand_dims(K.sum(x*K.expand_dims(self.kernel,axis = 1),axis =-2),1),3,-2),x*K.expand_dims(self.kernel,axis = -1)],-2)
        # return x*K.expand_dims(self.kernel,axis = -1)
        return K.sum(x*K.expand_dims(self.kernel,axis = 1),axis =-2)
    def compute_output_shape(self, input_shape):
        return input_shape[:-2] + (input_shape[-1],)





from elmoformanylangs import Embedder
import numpy as np

class plugin_elmo_many:

    def __init__(self):
        self.e = None
        self.dim = 0
        print("ELMo Plugin Loaded")

    def load(self, path):
        self.e = Embedder(path,batch_size=140)
        self.dim = (3,self.e.sents2elmo(["hello"],output_layer=3)[0].shape[-1])
        print("ELMo loaded")

    def getEmbedding(self,w):
        # elmo_emb = self.e.sents2elmo(w)
        # elmo_emb = np.concatenate([x for x in  self.e.sents2elmo([w],output_layer=3)[0]], axis=- 1)
        elmo_emb = np.swapaxes(self.e.sents2elmo([w], output_layer=3)[0],0,1)
        return(elmo_emb)
    def getDim(self):
        return self.dim
#
# p = plugin_elmo_many()
# p.load("/home/jborst/elmo/")
# e = p.getEmbedding(["hello", "world","german","embeddings"])
#
# e.shape


from raziel.plugins.plugin_fasttext import *
from raziel.plugins.plugin_elmo import *
from raziel.plugins.plugin_flair import *
from raziel.plugins.plugin_elmo_many import *
from raziel.plugins.plugin_chars import plugin_chars
from raziel.plugins.plugin_custom import plugin_custom
from raziel.plugins.plugin_label import plugin_label
from raziel.plugins.plugin_multilabel import plugin_multilabel
from raziel.plugins.plugin_list import plugin_list
from raziel.plugins.plugin_bert import plugin_bert
plugins_dict = { "fasttext": plugin_fasttext,
                 "flair":plugin_flair,
                 "elmo": plugin_elmo,
                 "character": plugin_chars,
                 "custom": plugin_custom,
                 "label":plugin_label,
                 "multilabel":plugin_multilabel,
                 "list": plugin_list,
                 "bert": plugin_bert}


import numpy as np
from raziel.vocab import Vocabulary


class plugin_multilabel:
    def __init__(self, level, option, **kwargs):
        self.vocab= Vocabulary()
        self.vocab.setTagset(kwargs["tagset"])
        self.name = "multilabel"
        self.level = level
        self.word_length = kwargs["word_length"]
        self.sentence_length = kwargs["sentence_length"]
        self.maximum_tagsetsize=self.vocab.numOfTags
        if kwargs.get("maximum_tagsetsize", None) is not None:
            self.maximum_tagsetsize = kwargs.get["maximum_tagsetsize"]

        assert level in ["index"], "only 'index' supported"

        self.option = "categorical"
        if option is not None:
            self.option = option
        assert option in ["sparse","categorical"] or None, \
            "Multilabel option not in sparse or categrorical"

        if option == "sparse":
            self.shape = (self.sentence_length, self.maximum_tagsetsize)
        elif option == "categorical":
            self.shape = (self.sentence_length, self.vocab.numOfTags)

        print("Multilabel plugin loaded")


    def getEmbedding(self, w ):
        index = [self.vocab.getTagsetIndex(x) for x in w]
        if self.option=="sparse":
            array = np.full((1,) + self.shape, 0., dtype="int32")
            for i, row in enumerate(index):
                array[0, i,:min(len(row),self.shape[1])] = row
            return array

        if self.option == "categorical":
            array = np.full((1,) + self.shape, 0., dtype="int32")
            for i, tags in enumerate(index):
                for tag in tags:
                    array[0, i, tag]=1
            return array


    def getDim(self):
        return self.shape



import numpy as np
from raziel.vocab import Vocabulary

class plugin_list:
    def __init__(self, level, option, **kwargs):
        self.vocab= Vocabulary()
        self.vocab.setTagset(option)
        self.name = "list"
        self.level = level
        self.word_length = kwargs["word_length"]
        self.sentence_length = kwargs["sentence_length"]
        assert level in ["index"], "only 'index' supported"

        self.option = "sparse"
        if option is not None:
            self.option = option
        self.shape = (self.sentence_length, )
        print("List plugin loaded")


    def getEmbedding(self, w ):
        index = self.vocab.getTagsetIndex(w)
        index_array = np.full((1, self.sentence_length), 0., dtype="int32")
        index_array[0, :min(self.sentence_length, len(index))] = index[:min(self.sentence_length, len(index))]
        return index_array


    def getDim(self):
        return (self.vocab.wordEmbeddingSize,)




import numpy as np
from raziel.vocab import Vocabulary
import keras

class plugin_chars:

    def __init__(self, level, option, **kwargs):
        self.vocab = Vocabulary()
        self.name = "characters"
        self.word_length = kwargs["word_length"]
        self.sentence_length = kwargs["sentence_length"]
        self.level= level
        assert level in ["index"], "%s needs 'index' as output. Embedding will be coming soon"
        if option is None:
            self.vocab.setCharacterset()
        else:
            assert isinstance(option, str), "Option for Characters must " \
                                            "be a string with all the characters"
            self.vocab.setCharacterset(option)

        if level == "index":
            self.shape = (self.sentence_length, self.word_length)

        print("Char Dict loaded")

    def getEmbedding(self, w):

        characters = np.full((1, self.sentence_length ,self.word_length), 0., dtype="float32")
        indices = keras.preprocessing.sequence.pad_sequences(
            [self.vocab.getAlphabetIndex(list(word)) for word in w], maxlen=self.word_length, padding="post", value=0.)

        characters[0,:min(self.sentence_length,indices.shape[0]),:] = indices
        return(characters)

    def getDim(self):
        return None



from keras.layers import LSTM
import keras.backend as K


class DropConnectLSTM(LSTM):
    def __init__(self, recurrent_dropconnect=0., dropconnect = 0., **kwargs):
        self.dropconnect = 1. - dropconnect
        self.recurrent_dropconnect = 1. - recurrent_dropconnect
        super(DropConnectLSTM, self).__init__(**kwargs)
        if 0. < self.dropconnect < 1. or 0. < self.recurrent_dropconnect < 1.:
            self.uses_learning_phase = True
        self.supports_masking = True

    def call(self, x, mask = None, **kwargs):
        if 0. < self.dropconnect < 1.:
            self.cell.kernel = K.in_train_phase(K.dropout(self.cell.kernel, self.dropconnect), self.cell.kernel)
            self.cell.bias = K.in_train_phase(K.dropout(self.cell.bias, self.dropconnect), self.cell.bias)
        if 0. < self.recurrent_dropconnect < 1.:
            self.cell.kernel = K.in_train_phase(K.dropout(self.cell.recurrent_kernel, self.recurrent_dropconnect), self.cell.kernel)
        return super(LSTM, self).call(x,mask,**kwargs)




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="raziel",
    version="0.0.3",
    author="Janos Borst",
    author_email="",
    description="A Data Transformation helper",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(), #['ptagger','ptagger.plugins'],#
    install_requires = ["allennlp", "flair", "numpy", "keras", "tqdm"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from __future__ import absolute_import
from __future__ import division

from keras import backend as K
from keras import activations
from keras import initializers
from keras import regularizers
from keras import constraints
from keras.engine import Layer
from keras.engine import InputSpec
from keras.objectives import categorical_crossentropy
from keras.objectives import sparse_categorical_crossentropy


class simpleCRF(Layer):
    """An implementation of linear chain conditional random field (CRF).

    An linear chain CRF is defined to maximize the following likelihood function:

    $$ L(W, U, b; y_1, ..., y_n) := \frac{1}{Z} \sum_{y_1, ..., y_n} \exp(-a_1' y_1 - a_n' y_n
        - \sum_{k=1^n}((f(x_k' W + b) y_k) + y_1' U y_2)), $$

    where:
        $Z$: normalization constant
        $x_k, y_k$:  inputs and outputs

    This implementation has two modes for optimization:
    1. (`join mode`) optimized by maximizing join likelihood, which is optimal in theory of statistics.
       Note that in this case, CRF must be the output/last layer.
    2. (`marginal mode`) return marginal probabilities on each time step and optimized via composition
       likelihood (product of marginal likelihood), i.e., using `categorical_crossentropy` loss.
       Note that in this case, CRF can be either the last layer or an intermediate layer (though not explored).

    For prediction (test phrase), one can choose either Viterbi best path (class indices) or marginal
    probabilities if probabilities are needed. However, if one chooses *join mode* for training,
    Viterbi output is typically better than marginal output, but the marginal output will still perform
    reasonably close, while if *marginal mode* is used for training, marginal output usually performs
    much better. The default behavior is set according to this observation.

    In addition, this implementation supports masking and accepts either onehot or sparse target.


    # Examples

    ```python
        model = Sequential()
        model.add(Embedding(3001, 300, mask_zero=True)(X)

        # use learn_mode = 'join', test_mode = 'viterbi', sparse_target = True (label indice output)
        crf = CRF(10, sparse_target=True)
        model.add(crf)

        # crf.accuracy is default to Viterbi acc if using join-mode (default).
        # One can add crf.marginal_acc if interested, but may slow down learning
        model.compile('adam', loss=crf.loss_function, metrics=[crf.accuracy])

        # y must be label indices (with shape 1 at dim 3) here, since `sparse_target=True`
        model.fit(x, y)

        # prediction give onehot representation of Viterbi best path
        y_hat = model.predict(x_test)
    ```


    # Arguments
        units: Positive integer, dimensionality of the output space.
        learn_mode: Either 'join' or 'marginal'.
            The former train the model by maximizing join likelihood while the latter
            maximize the product of marginal likelihood over all time steps.
        test_mode: Either 'viterbi' or 'marginal'.
            The former is recommended and as default when `learn_mode = 'join'` and
            gives one-hot representation of the best path at test (prediction) time,
            while the latter is recommended and chosen as default when `learn_mode = 'marginal'`,
            which produces marginal probabilities for each time step.
        sparse_target: Boolean (default False) indicating if provided labels are one-hot or
            indices (with shape 1 at dim 3).
        use_boundary: Boolean (default True) indicating if trainable start-end chain energies
            should be added to model.
        use_bias: Boolean, whether the layer uses a bias vector.
        kernel_initializer: Initializer for the `kernel` weights matrix,
            used for the linear transformation of the inputs.
            (see [initializers](../initializers.md)).
        chain_initializer: Initializer for the `chain_kernel` weights matrix,
            used for the CRF chain energy.
            (see [initializers](../initializers.md)).
        boundary_initializer: Initializer for the `left_boundary`, 'right_boundary' weights vectors,
            used for the start/left and end/right boundary energy.
            (see [initializers](../initializers.md)).
        bias_initializer: Initializer for the bias vector
            (see [initializers](../initializers.md)).
        activation: Activation function to use
            (see [activations](../activations.md)).
            If you pass None, no activation is applied
            (ie. "linear" activation: `a(x) = x`).
        kernel_regularizer: Regularizer function applied to
            the `kernel` weights matrix
            (see [regularizer](../regularizers.md)).
        chain_regularizer: Regularizer function applied to
            the `chain_kernel` weights matrix
            (see [regularizer](../regularizers.md)).
        boundary_regularizer: Regularizer function applied to
            the 'left_boundary', 'right_boundary' weight vectors
            (see [regularizer](../regularizers.md)).
        bias_regularizer: Regularizer function applied to the bias vector
            (see [regularizer](../regularizers.md)).
        kernel_constraint: Constraint function applied to
            the `kernel` weights matrix
            (see [constraints](../constraints.md)).
        chain_constraint: Constraint function applied to
            the `chain_kernel` weights matrix
            (see [constraints](../constraints.md)).
        boundary_constraint: Constraint function applied to
            the `left_boundary`, `right_boundary` weights vectors
            (see [constraints](../constraints.md)).
        bias_constraint: Constraint function applied to the bias vector
            (see [constraints](../constraints.md)).
        input_dim: dimensionality of the input (integer).
            This argument (or alternatively, the keyword argument `input_shape`)
            is required when using this layer as the first layer in a model.
        unroll: Boolean (default False). If True, the network will be unrolled, else a symbolic loop will be used.
            Unrolling can speed-up a RNN, although it tends to be more memory-intensive.
            Unrolling is only suitable for short sequences.

    # Input shape
        3D tensor with shape `(nb_samples, timesteps, input_dim)`.

    # Output shape
        3D tensor with shape `(nb_samples, timesteps, units)`.

    # Masking
        This layer supports masking for input data with a variable number
        of timesteps. To introduce masks to your data,
        use an [Embedding](embeddings.md) layer with the `mask_zero` parameter
        set to `True`.

    """

    def __init__(self, units,
                 learn_mode='join',
                 test_mode=None,
                 sparse_target=False,
                 use_boundary=True,
                 use_bias=True,
                 activation='linear',
                 kernel_initializer='glorot_uniform',
                 chain_initializer='orthogonal',
                 bias_initializer='zeros',
                 boundary_initializer='zeros',
                 kernel_regularizer=None,
                 chain_regularizer=None,
                 boundary_regularizer=None,
                 bias_regularizer=None,
                 kernel_constraint=None,
                 chain_constraint=None,
                 boundary_constraint=None,
                 bias_constraint=None,
                 input_dim=None,
                 unroll=False,
                 **kwargs):
        super(simpleCRF, self).__init__(**kwargs)
        self.supports_masking = True
        self.units = units
        self.learn_mode = learn_mode
        assert self.learn_mode in ['join', 'marginal']
        self.test_mode = test_mode
        if self.test_mode is None:
            self.test_mode = 'viterbi' if self.learn_mode == 'join' else 'marginal'
        else:
            assert self.test_mode in ['viterbi', 'marginal']
        self.sparse_target = sparse_target
        self.use_boundary = use_boundary
        self.use_bias = use_bias

        self.activation = activations.get(activation)

        self.chain_initializer = initializers.get(chain_initializer)
        self.boundary_initializer = initializers.get(boundary_initializer)
        self.bias_initializer = initializers.get(bias_initializer)

        self.chain_regularizer = regularizers.get(chain_regularizer)
        self.boundary_regularizer = regularizers.get(boundary_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)

        self.chain_constraint = constraints.get(chain_constraint)
        self.boundary_constraint = constraints.get(boundary_constraint)
        self.bias_constraint = constraints.get(bias_constraint)

        self.unroll = unroll

    def build(self, input_shape):
        self.input_spec = [InputSpec(shape=input_shape)]
        self.input_dim = input_shape[-1]


        self.chain_kernel = self.add_weight((self.units, self.units),
                                            name='chain_kernel',
                                            initializer=self.chain_initializer,
                                            regularizer=self.chain_regularizer,
                                            constraint=self.chain_constraint)
        if self.use_boundary:
            self.left_boundary = self.add_weight((self.units,),
                                                 name='left_boundary',
                                                 initializer=self.boundary_initializer,
                                                 regularizer=self.boundary_regularizer,
                                                 constraint=self.boundary_constraint)
            self.right_boundary = self.add_weight((self.units,),
                                                  name='right_boundary',
                                                  initializer=self.boundary_initializer,
                                                  regularizer=self.boundary_regularizer,
                                                  constraint=self.boundary_constraint)
        self.built = True

    def call(self, X, mask=None):
        if mask is not None:
            assert K.ndim(mask) == 2, 'Input mask to CRF must have dim 2 if not None'

        if self.test_mode == 'viterbi':
            test_output = self.viterbi_decoding(X, mask)
        else:
            test_output = self.get_marginal_prob(X, mask)

        self.uses_learning_phase = True
        if self.learn_mode == 'join':
            train_output = K.zeros_like(X)
            out = K.in_train_phase(train_output, test_output)
        else:
            if self.test_mode == 'viterbi':
                train_output = self.get_marginal_prob(X, mask)
                out = K.in_train_phase(train_output, test_output)
            else:
                out = test_output
        return out

    def compute_output_shape(self, input_shape):
        return input_shape

    def compute_mask(self, input, mask=None):
        if mask is not None and self.learn_mode == 'join':
            return K.any(mask, axis=1)
        return mask

    def get_config(self):
        config = {'units': self.units,
                  'learn_mode': self.learn_mode,
                  'test_mode': self.test_mode,
                  'use_boundary': self.use_boundary,
                  'use_bias': self.use_bias,
                  'sparse_target': self.sparse_target,
                  'chain_initializer': initializers.serialize(self.chain_initializer),
                  'boundary_initializer': initializers.serialize(self.boundary_initializer),
                  'bias_initializer': initializers.serialize(self.bias_initializer),
                  'activation': activations.serialize(self.activation),
                  'chain_regularizer': regularizers.serialize(self.chain_regularizer),
                  'boundary_regularizer': regularizers.serialize(self.boundary_regularizer),
                  'bias_regularizer': regularizers.serialize(self.bias_regularizer),
                  'chain_constraint': constraints.serialize(self.chain_constraint),
                  'boundary_constraint': constraints.serialize(self.boundary_constraint),
                  'bias_constraint': constraints.serialize(self.bias_constraint),
                  'input_dim': self.input_dim,
                  'unroll': self.unroll}
        base_config = super(simpleCRF, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    @property
    def loss_function(self):
        if self.learn_mode == 'join':
            def loss(y_true, y_pred):
                assert self._inbound_nodes, 'CRF has not connected to any layer.'
                assert not self._outbound_nodes, 'When learn_model="join", CRF must be the last layer.'
                if self.sparse_target:
                    y_true = K.one_hot(K.cast(y_true[:, :, 0], 'int32'), self.units)
                X = self._inbound_nodes[0].input_tensors[0]
                mask = self._inbound_nodes[0].input_masks[0]
                nloglik = self.get_negative_log_likelihood(y_true, X, mask)
                return nloglik
            return loss
        else:
            if self.sparse_target:
                return sparse_categorical_crossentropy
            else:
                return categorical_crossentropy

    @property
    def accuracy(self):
        if self.test_mode == 'viterbi':
            return self.viterbi_acc
        else:
            return self.marginal_acc

    @staticmethod
    def _get_accuracy(y_true, y_pred, mask, sparse_target=False):
        y_pred = K.argmax(y_pred, -1)
        if sparse_target:
            y_true = K.cast(y_true[:, :, 0], K.dtype(y_pred))
        else:
            y_true = K.argmax(y_true, -1)
        judge = K.cast(K.equal(y_pred, y_true), K.floatx())
        if mask is None:
            return K.mean(judge)
        else:
            mask = K.cast(mask, K.floatx())
            return K.sum(judge * mask) / K.sum(mask)

    @property
    def viterbi_acc(self):
        def acc(y_true, y_pred):
            X = self._inbound_nodes[0].input_tensors[0]
            mask = self._inbound_nodes[0].input_masks[0]
            y_pred = self.viterbi_decoding(X, mask)
            return self._get_accuracy(y_true, y_pred, mask, self.sparse_target)
        acc.func_name = 'viterbi_acc'
        return acc

    @property
    def marginal_acc(self):
        def acc(y_true, y_pred):
            X = self._inbound_nodes[0].input_tensors[0]
            mask = self._inbound_nodes[0].input_masks[0]
            y_pred = self.get_marginal_prob(X, mask)
            return self._get_accuracy(y_true, y_pred, mask, self.sparse_target)
        acc.func_name = 'marginal_acc'
        return acc

    @staticmethod
    def softmaxNd(x, axis=-1):
        m = K.max(x, axis=axis, keepdims=True)
        exp_x = K.exp(x - m)
        prob_x = exp_x / K.sum(exp_x, axis=axis, keepdims=True)
        return prob_x

    @staticmethod
    def shift_left(x, offset=1):
        assert offset > 0
        return K.concatenate([x[:, offset:], K.zeros_like(x[:, :offset])], axis=1)

    @staticmethod
    def shift_right(x, offset=1):
        assert offset > 0
        return K.concatenate([K.zeros_like(x[:, :offset]), x[:, :-offset]], axis=1)

    def add_boundary_energy(self, energy, mask, start, end):
        start = K.expand_dims(K.expand_dims(start, 0), 0)
        end = K.expand_dims(K.expand_dims(end, 0), 0)
        if mask is None:
            energy = K.concatenate([energy[:, :1, :] + start, energy[:, 1:, :]], axis=1)
            energy = K.concatenate([energy[:, :-1, :], energy[:, -1:, :] + end], axis=1)
        else:
            mask = K.expand_dims(K.cast(mask, K.floatx()))
            start_mask = K.cast(K.greater(mask, self.shift_right(mask)), K.floatx())
            end_mask = K.cast(K.greater(self.shift_left(mask), mask), K.floatx())
            energy = energy + start_mask * start
            energy = energy + end_mask * end
        return energy

    def get_log_normalization_constant(self, input_energy, mask, **kwargs):
        """Compute logarithm of the normalization constant Z, where
        Z = sum exp(-E) -> logZ = log sum exp(-E) =: -nlogZ
        """
        # should have logZ[:, i] == logZ[:, j] for any i, j
        logZ = self.recursion(input_energy, mask, return_sequences=False, **kwargs)
        return logZ[:, 0]

    def get_energy(self, y_true, input_energy, mask):
        """Energy = a1' y1 + u1' y1 + y1' U y2 + u2' y2 + y2' U y3 + u3' y3 + an' y3
        """
        input_energy = K.sum(input_energy * y_true, 2)  # (B, T)
        chain_energy = K.sum(K.dot(y_true[:, :-1, :], self.chain_kernel) * y_true[:, 1:, :], 2)  # (B, T-1)

        if mask is not None:
            mask = K.cast(mask, K.floatx())
            chain_mask = mask[:, :-1] * mask[:, 1:]  # (B, T-1), mask[:,:-1]*mask[:,1:] makes it work with any padding
            input_energy = input_energy * mask
            chain_energy = chain_energy * chain_mask
        total_energy = K.sum(input_energy, -1) + K.sum(chain_energy, -1)  # (B, )

        return total_energy

    def get_negative_log_likelihood(self, y_true, X, mask):
        """Compute the loss, i.e., negative log likelihood (normalize by number of time steps)
           likelihood = 1/Z * exp(-E) ->  neg_log_like = - log(1/Z * exp(-E)) = logZ + E
        """
        input_energy = X
        if self.use_boundary:
            input_energy = self.add_boundary_energy(input_energy, mask, self.left_boundary, self.right_boundary)
        energy = self.get_energy(y_true, input_energy, mask)
        logZ = self.get_log_normalization_constant(input_energy, mask, input_length=K.int_shape(X)[1])
        nloglik = logZ + energy
        if mask is not None:
            nloglik = nloglik / K.sum(K.cast(mask, K.floatx()), 1)
        else:
            nloglik = nloglik / K.cast(K.shape(X)[1], K.floatx())
        return nloglik

    def step(self, input_energy_t, states, return_logZ=True):
        # not in the following  `prev_target_val` has shape = (B, F)
        # where B = batch_size, F = output feature dim
        # Note: `i` is of float32, due to the behavior of `K.rnn`
        prev_target_val, i, chain_energy = states[:3]
        t = K.cast(i[0, 0], dtype='int32')
        if len(states) > 3:
            if K.backend() == 'theano':
                m = states[3][:, t:(t + 2)]
            else:
                m = K.tf.slice(states[3], [0, t], [-1, 2])
            input_energy_t = input_energy_t * K.expand_dims(m[:, 0])
            chain_energy = chain_energy * K.expand_dims(K.expand_dims(m[:, 0] * m[:, 1]))  # (1, F, F)*(B, 1, 1) -> (B, F, F)
        if return_logZ:
            energy = chain_energy + K.expand_dims(input_energy_t - prev_target_val, 2)  # shapes: (1, B, F) + (B, F, 1) -> (B, F, F)
            new_target_val = K.logsumexp(-energy, 1)  # shapes: (B, F)
            return new_target_val, [new_target_val, i + 1]
        else:
            energy = chain_energy + K.expand_dims(input_energy_t + prev_target_val, 2)
            min_energy = K.min(energy, 1)
            argmin_table = K.cast(K.argmin(energy, 1), K.floatx())  # cast for tf-version `K.rnn`
            return argmin_table, [min_energy, i + 1]

    def recursion(self, input_energy, mask=None, go_backwards=False, return_sequences=True, return_logZ=True, input_length=None):
        """Forward (alpha) or backward (beta) recursion

        If `return_logZ = True`, compute the logZ, the normalization constant:

        \[ Z = \sum_{y1, y2, y3} exp(-E) # energy
          = \sum_{y1, y2, y3} exp(-(u1' y1 + y1' W y2 + u2' y2 + y2' W y3 + u3' y3))
          = sum_{y2, y3} (exp(-(u2' y2 + y2' W y3 + u3' y3)) sum_{y1} exp(-(u1' y1' + y1' W y2))) \]

        Denote:
            \[ S(y2) := sum_{y1} exp(-(u1' y1 + y1' W y2)), \]
            \[ Z = sum_{y2, y3} exp(log S(y2) - (u2' y2 + y2' W y3 + u3' y3)) \]
            \[ logS(y2) = log S(y2) = log_sum_exp(-(u1' y1' + y1' W y2)) \]
        Note that:
              yi's are one-hot vectors
              u1, u3: boundary energies have been merged

        If `return_logZ = False`, compute the Viterbi's best path lookup table.
        """
        chain_energy = self.chain_kernel
        chain_energy = K.expand_dims(chain_energy, 0)  # shape=(1, F, F): F=num of output features. 1st F is for t-1, 2nd F for t
        prev_target_val = K.zeros_like(input_energy[:, 0, :])  # shape=(B, F), dtype=float32

        if go_backwards:
            input_energy = K.reverse(input_energy, 1)
            if mask is not None:
                mask = K.reverse(mask, 1)

        initial_states = [prev_target_val, K.zeros_like(prev_target_val[:, :1])]
        constants = [chain_energy]

        if mask is not None:
            mask2 = K.cast(K.concatenate([mask, K.zeros_like(mask[:, :1])], axis=1), K.floatx())
            constants.append(mask2)

        def _step(input_energy_i, states):
            return self.step(input_energy_i, states, return_logZ)

        target_val_last, target_val_seq, _ = K.rnn(_step, input_energy, initial_states, constants=constants,
                                                   input_length=input_length, unroll=self.unroll)

        if return_sequences:
            if go_backwards:
                target_val_seq = K.reverse(target_val_seq, 1)
            return target_val_seq
        else:
            return target_val_last

    def forward_recursion(self, input_energy, **kwargs):
        return self.recursion(input_energy, **kwargs)

    def backward_recursion(self, input_energy, **kwargs):
        return self.recursion(input_energy, go_backwards=True, **kwargs)

    def get_marginal_prob(self, X, mask=None):
        input_energy = X
        if self.use_boundary:
            input_energy = self.add_boundary_energy(input_energy, mask, self.left_boundary, self.right_boundary)
        input_length = K.int_shape(X)[1]
        alpha = self.forward_recursion(input_energy, mask=mask, input_length=input_length)
        beta = self.backward_recursion(input_energy, mask=mask, input_length=input_length)
        if mask is not None:
            input_energy = input_energy * K.expand_dims(K.cast(mask, K.floatx()))
        margin = -(self.shift_right(alpha) + input_energy + self.shift_left(beta))
        return self.softmaxNd(margin)

    def viterbi_decoding(self, X, mask=None):
        input_energy = X
        if self.use_boundary:
            input_energy = self.add_boundary_energy(input_energy, mask, self.left_boundary, self.right_boundary)

        argmin_tables = self.recursion(input_energy, mask, return_logZ=False)
        argmin_tables = K.cast(argmin_tables, 'int32')

        # backward to find best path, `initial_best_idx` can be any, as all elements in the last argmin_table are the same
        argmin_tables = K.reverse(argmin_tables, 1)
        initial_best_idx = [K.expand_dims(argmin_tables[:, 0, 0])]  # matrix instead of vector is required by tf `K.rnn`
        if K.backend() == 'theano':
            initial_best_idx = [K.T.unbroadcast(initial_best_idx[0], 1)]

        def gather_each_row(params, indices):
            n = K.shape(indices)[0]
            if K.backend() == 'theano':
                return params[K.T.arange(n), indices]
            else:
                indices = K.transpose(K.stack([K.tf.range(n), indices]))
                return K.tf.gather_nd(params, indices)

        def find_path(argmin_table, best_idx):
            next_best_idx = gather_each_row(argmin_table, best_idx[0][:, 0])
            next_best_idx = K.expand_dims(next_best_idx)
            if K.backend() == 'theano':
                next_best_idx = K.T.unbroadcast(next_best_idx, 1)
            return next_best_idx, [next_best_idx]

        _, best_paths, _ = K.rnn(find_path, argmin_tables, initial_best_idx, input_length=K.int_shape(X)[1], unroll=self.unroll)
        best_paths = K.reverse(best_paths, 1)
        best_paths = K.squeeze(best_paths, 2)

        return K.one_hot(best_paths, self.units)


import string
import numpy as np
import logging
# from . import helpers



class Vocabulary:
	def __init__(self):
		self.data = None
		self.specialToken = "UNK"
		self.lowercase=False

		self.alphabet = list(string.ascii_letters+"0123456789"+string.punctuation+"äöüÄÖÜ \n")
		self.wordList = None
		self.tagset = None

		self.lookupAlphabet = None
		self.lookupWords = None
		self.lookupWordsReverse = None
		self.lookupTagset = None
		self.lookupTagsetReverse = None

		self.wordEmbedding=None
		self.wordEmbeddingSize = None
		self.vocabularySize = 0
		self.alphabetSize = None
		self.numOfTags = None


		self.conllData = None

		self.logger = logging.getLogger(self.__class__.__name__)
		self.logger.setLevel(logging.WARNING)

	###################################################
	# Setter

	def setCharacterset(self, alphabet = None):
		self.logger.debug("Setting alphabet...")
		if alphabet:
			self.alphabet = list(alphabet)
		self.alphabetSize = len(self.alphabet) + 1
		self.logger.debug("Setting alphabetsize to %i" % (self.alphabetSize,))
		self.lookupAlphabet = dict(zip(self.alphabet,list(range(1,self.alphabetSize))))
		self.lookupAlphabetReverse = dict(zip(list(range(1, self.alphabetSize)),self.alphabet))

	def setTagset(self, tagset = None):
		if not tagset:
			self.logging.error("Tagset empty in self.setTagset()...")
			exit()
		self.tagset=tagset
		self.numOfTags = len(tagset)
		self.logger.debug("Setting tagsetsize to %i" % (self.numOfTags,))
		self.lookupTagset =dict(zip(self.tagset,list(range(0,self.numOfTags))))
		self.logger.debug("Creating reverse lookups for tags...")
		self.lookupTagsetReverse =dict(zip(list(range(0,self.numOfTags)) ,self.tagset))

	def setWordEmbedding(self, path=None):
		self.logger.debug("Loading static word embedding")
		from gensim.models import KeyedVectors as kv
		self.logger.debug("Trying " + path)
		self.wordEmbedding = kv.load_word2vec_format(path)

		self.logger.debug("Loaded %i vectors " % (len(self.wordEmbedding.index2word),) )
		self.logger.debug("Stacking 0 and special token ")
		self.wordEmbedding.vectors = np.vstack((
			np.array([0]* self.wordEmbedding.vector_size), # the vector for the masking
			self.wordEmbedding.vectors,
			np.array([0] * self.wordEmbedding.vector_size),  # the vector for the special token
		))

		self.wordEmbedding.index2word.insert(0,"")
		self.wordEmbedding.index2word.append(self.specialToken)


		self.vocabularySize = len(self.wordEmbedding.index2word)-1
		self.logger.debug("Setting vocabularysize to %i . (Loaded vectors + special token)" % (self.vocabularySize,))
		self.wordEmbeddingSize = self.wordEmbedding.vector_size

		self.wordList = self.wordEmbedding.index2word[1:  len(self.wordEmbedding.index2word)+1]
		self.logger.debug("Creating lookup..")
		self.lookupWords = dict(zip(self.wordList,list(range(1,self.vocabularySize +1))))

	################################################################
	# Getters
	def  getEmbeddingLayer(self,trainable = False):
		"Get the loaded pretrained embedding as a keras layer, ready to be imported to the neural network"
		if self.wordEmbedding:
			return (self.wordEmbedding.get_keras_embedding(train_embeddings=trainable))
		else:
			self.logger.warning("Trying to get a word embedding layer with no static embedding loaded.")
			return(None)

	def getWordIndex(self, tokens = None):
		if type(tokens) == list :
			if self.lowercase:
				return([self.lookupWords.get(item.lower(),self.vocabularySize) for item in tokens])
			else:
				return([self.lookupWords.get(item,self.vocabularySize) for item in tokens])

		elif type(tokens) == str:
			if self.lowercase:
				return(self.lookupWords.get(tokens.lower(),self.vocabularySize))
			else:
				return(self.lookupWords.get(tokens,self.vocabularySize))
	def getWordIndexReverse(self, tokens=None):
		if self.lookupWordsReverse is None:
			self.lookupWordsReverse = dict(zip( list(self.lookupWords.values()),list(self.lookupWords.keys()) ))
		if type(tokens) == list:
			return ([self.lookupWordsReverse.get(item, 'unk') for item in tokens])
		else:
			return (self.lookupWordsReverse.get(tokens, 'unk'))



	def getAlphabetIndex(self, tokens=None):
		if type(tokens) == list:
			return ([self.lookupAlphabet.get(item, self.alphabetSize) for item in tokens])
		elif type(tokens) == str:
			return (self.lookupAlphabet.get(tokens, self.alphabetSize))

	def getAlphabetIndexReverse(self, tag=None):
		if type(tag) == list:
			return ([self.lookupAlphabetReverse.get(item, 'unk') for item in tag])
		else:
			return (self.lookupAlphabetReverse.get(tag, 'unk'))

	def getTagsetIndex(self, tag=None):
		if type(tag) == list:
			return ([self.lookupTagset[item] for item in tag])
		elif type(tag) == str:
			return (self.lookupTagset[tag])

	def getTagsetIndexReverse(self, tag=None):
		if type(tag) == list:
			return ([self.lookupTagsetReverse.get(item, None) for item in tag])
		else:
			return (self.lookupTagsetReverse.get(tag, None))


	def __getstate__(self):
		blacklist = ["logger", "wordEmbedding"]
		print("Static loaded word Embeddings won't be saved outside the model!")
		return {key: self.__dict__[key] for key in self.__dict__.keys() if key not in blacklist}
	def __setstate__(self, state):
		for key in state.keys():
			self.__dict__[key] = state[key]
			self.logger = logging.getLogger(self.__class__.__name__)
			self.logger.setLevel(logging.WARNING)




def main():
	vocab = Vocabulary()
	vocab.readConll("/home/jb/git/ptagger/resources/BIOES/test_small2.txt")
	vocab.setWordEmbedding("/home/jb/git/ptagger/resources/cc.de.300.10K.vec")
	# vocab.getWordIndex("Parlament")
	vocab.setCharacterset()
	vocab.setTagset(["O","S-PER","S-LOC","S-ORG","S-MISC","B-PER","B-LOC","B-ORG","B-MISC","I-PER","I-LOC","I-ORG","I-MISC","E-PER","E-LOC","E-ORG","E-MISC"])
	#
	#
	trainingsdata = vocab.getTraindata(get = ["words","chars","tags"])

if __name__ == "__main__":
	main()


import numpy as np
import os
import pathlib
import logging
from raziel.plugins import plugins_dict
from raziel.vocab import Vocabulary
from tqdm import tqdm





class Raziel:
    def __init__(self, dm = None):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(logging.DEBUG)

        self.vocab = Vocabulary()
        self.externalPlugins = {}

        self.charVocabularysize = None
        self.max_word_length = 60
        self.max_sequence_length = 140

        self.pluginDimension = {}
        self.numOfTags = None


        self.scheme = ""
        self.defaultDirectory = os.path.join(pathlib.Path.home().as_posix(), ".raziel")

        self.data_manager = dm
        self.transform_chain= None


    def setLogging(self,level):
        if level.lower() == "debug":
            self.logger.setLevel(logging.DEBUG)
        if level.lower() == "error":
            self.logger.setLevel(logging.ERROR)
        if level.lower() == "warning":
            self.logger.setLevel(logging.WARNING)

    def load_plugins(self, params):
        assert isinstance(params, dict)
        self.pluginDimension={}
        self.externalPlugins={}
        for key, val in params.items():
            self.externalPlugins[key] = plugins_dict[key]()
            self.externalPlugins[key].load(val)
            self.pluginDimension[key] = self.externalPlugins[key].getDim()
            if self.wordEmbeddingDimension is None:
                Warning("Word embedding dimension not set.")


    def toScheme (self, tagset, scheme="iobes", outside="O"):
        scheme = scheme.lower()
        if scheme == "iobes" or scheme == "bioes":
            return ([outside]+["I-"+x for x in tagset]+["B-"+x for x in tagset]+["E-"+x for x in tagset]+["S-"+x for x in tagset])
        elif (scheme == "bilou"):
            return (
            [outside] + ["B-" + x for x in tagset] + ["I-" + x for x in tagset] + ["L-" + x for x in tagset] + ["U-" + x for x in tagset])
        elif (scheme == "bio" or scheme == "iob"):
            return ([outside] + ["B-" + x for x in tagset] + ["I-" + x for x in tagset])
        elif(scheme == "noprefix"):
            return [outside] + tagset

        return(tagset)

    def setTagset(self, tagset, scheme="iobes", outside="O"):
        self.scheme = scheme
        self.vocab.setTagset(self.toScheme(tagset, scheme=scheme, outside=outside))


    def chain(self, get):
        #print (get)
        self.transform_chain = {}
        kwargs = {
            "sentence_length": self.max_sequence_length,
            "word_length": self.max_word_length,
            "tagset": self.vocab.tagset
        }
        for name, description in get.items():
            self.transform_chain[name] = (description[0], plugins_dict[description[1]](level=description[2], option=description[3], **kwargs))

    def transform(self, purpose = "general"):
        data = self.data_manager[purpose]
        traindata = {}
        traindata["lengths"] = data.groupby("sentence_id").count()["token"].tolist()
        traindata["raw"] = []
        print("helloe")
        dataset_size = len(set(data["sentence_id"].tolist()))
        for name, (column, plugin) in self.transform_chain.items():
            array = np.full((dataset_size,) + plugin.shape, 0.)
            print(name)
            for i, sentence in tqdm(enumerate(data.groupby("sentence_id"))):
                traindata["raw"].append(sentence[1]["token"].tolist())
                array[i] = plugin.getEmbedding(sentence[1][column].tolist())
            traindata[name]=array
        return traindata

    def transform_iter(self, batch_size=100,purpose = "general"):
        for data in self.data_manager.iter(batch_size=batch_size, purpose=purpose):
            traindata = {}
            traindata["lengths"] = np.array(data.groupby("sentence_id").count()["token"].tolist(),dtype=np.int32)
            traindata["sentence_ids"] = list(set(data["sentence_id"].tolist()))
            for name, (column, plugin) in self.transform_chain.items():
                array = np.full((batch_size,) + plugin.shape, 0.)
                print(name)
                for i, sentence in tqdm(enumerate(data.groupby("sentence_id"))):
                    array[i] = plugin.getEmbedding(sentence[1][column].tolist())
                traindata[name]=array
            yield traindata

    def getEmbeddingLayer(self):
        layers = {}
        for name, (column, plugin) in self.transform_chain.items():
            if hasattr(plugin, "vocab"):
                if hasattr(plugin.vocab, "wordEmbedding") and plugin.vocab.wordEmbedding is not None:
                    layers[name] = plugin.vocab.getEmbeddingLayer()
        return layers

    def schemetransformer(self, column, scheme="BIOES", multilabel=False, purpose="general"):
        scheme = scheme.upper()
        scheme= "BIOES" if scheme =="IOBES" else scheme
        data = self.data_manager.get_data(purpose=purpose)
        transforming_column = [tagset if isinstance(tagset,list) else [tagset] for tagset in data[column].tolist()]

        newtags_column = []
        for i, tagset in enumerate(transforming_column):
            if "O" in tagset or "O" == tagset:
                newtags_column.append(tagset)
            elif scheme == "NOPREFIX":
                newtags_column.append([t.split("-")[-1] for t in tagset])
            else:
                currentTagset = [t.split("-")[-1] for t in tagset]
                newtags = []
                if i == 0:
                    nextTagset = [t.split("-")[-1] for t in transforming_column[i + 1]]
                    for t in currentTagset:
                        if t in nextTagset:
                            newtags.append("B-" + t)
                        else:
                            if scheme == "BIOES": newtags.append("S-" + t)
                            if scheme == "BIO1": newtags.append("I-" + t)
                            if scheme == "BIO2": newtags.append("B-" + t)
                            newtags.append(newtags)
                elif i == len(transforming_column)-1:
                    previousTagset = [t.split("-")[-1] for t in
                                  transforming_column[i - 1]]
                    for t in currentTagset:
                        if t in previousTagset:
                            if scheme == "BIOES": newtags.append("E-" + t)
                            if scheme == "BIO1": newtags.append("I-" + t)
                            if scheme == "BIO2": newtags.append("I-" + t)
                        else:
                            if scheme == "BIOES": newtags.append("S-" + t)
                            if scheme == "BIO1": newtags.append("I-" + t)
                            if scheme == "BIO2": newtags.append("B-" + t)
                    newtags_column.append(newtags)
                else:

                    nextTagset = [t.split("-")[-1] for t in
                                  transforming_column[i + 1]]
                    previousTagset = [t.split("-")[-1] for t in
                                  transforming_column[i - 1]]
                    newtags = []
                    for t in currentTagset:
                        if t in previousTagset and t in nextTagset:
                            newtags.append("I-" + t)
                        elif t in previousTagset and t not in nextTagset:
                            if scheme == "BIOES": newtags.append("E-" + t)
                            if scheme == "BIO1": newtags.append("I-" + t)
                            if scheme == "BIO2": newtags.append("I-" + t)
                        elif t not in previousTagset and t in nextTagset:
                            newtags.append("B-" + t)
                        else:
                            if scheme == "BIOES": newtags.append("S-" + t)
                            if scheme == "BIO1": newtags.append("I-" + t)
                            if scheme == "BIO2": newtags.append("B-" + t)
                    newtags_column.append(newtags)

        if not multilabel:
            newtags_column = [x[0] for x in newtags_column]
        self.data_manager.update({column: newtags_column}, level="token")

def main():
    m = "cnn_features"



if __name__ == "__main__":
    main()